( 3iI2C.fs - AE - 2021-06-03---Thu---14:24 )
( GForth file for communicating directly with OLED display's SSD1306 driver IC, allowing for bare metal development, 
  i.e. without use of Adafruit's SSD1306 device driver software or their GFX graphics library. )
\ publication mention on Hackaday :)
\ https://hackaday.com/2021/07/19/an-oled-photo-frame-powered-by-the-attiny85/

\ TODO: Embed fixed-width Font into OLED driver and embedded font display for fast draw

\ exported functions from 3iI2C.avr -- these use Wire.h for I2C comms
$1F constant CALL_I2C_INIT
$20 constant CALL_I2C_SETCLOCK
$21 constant CALL_I2C_OLED1306
$22 constant CALL_I2C_OLEDFILL
$23 constant CALL_I2C_OLEDIMAGE
$24 constant CALL_I2C_OLEDINIT
$25 constant CALL_I2C_OLEDCHAR

\ register location to store I2C OLED display address - used by I2C_OLED1306
31 constant OLED_I2C_REGISTER
$3C constant OLED_I2C_ADDRESS  \ default address for SSD1306 based cheap OLEDs or $3D if set SMT resistor on back of board
\ TODO: what about read/write

\ SSD1306 commands 
$0  constant  SSD1306_COMMAND_FLAG
$40 constant  SSD1306_DATA_FLAG

: i2c-init ( -- )   \ Wire begin - start I2C
    CALL_I2C_INIT xcall ;

: i2c-setclock ( val/10k -- ) \ e.g. 10 means 100kHz bitrate   \  TODO: see if faster clock (100, 800) means snappier display
    xstore CALL_I2C_SETCLOCK xcall ;

: i2c-restoreclock ( val/10k -- ) \ e.g. 10 means 100kHz bitrate
    xstore CALL_I2C_SETCLOCK xcall ;
\ TODO: CODE REVIEW - what timing should be restored to?

: oled-setreg ( i2c-adr -- ) \ store I2C write address into fixed register
            \ ex: $3C OLED-display with SSD1306 driver IC - note: address changes when switching between write/read
   OLED_I2C_REGISTER setreg  ;

\ CODE REVIEW: consider to use nstore instead of xstore for communication efficiency!       
: [oled-bytes]  ( b1 b2 b3 ... bn N -- )  \ sends N bytes in reverse order and puts N at the end (target TOS)
    >r r@ revstore 
    r> xstore ;
: [oled-send]   >r [oled-bytes] r> xstore CALL_I2C_OLED1306 xcall ;
: oled-cmd  ( c1 c2 ... cn N -- )  SSD1306_COMMAND_FLAG  [oled-send] ;
: oled-data ( d1 d2 ... dn N -- )  SSD1306_DATA_FLAG     [oled-send] ;

\ ===================================================
\ SSD1306 OLED driver IC commands -- see datasheet
$AE  constant  SSD1306_DISPLAYOFF      \ default, sleep, data not affected
$AF  constant  SSD1306_DISPLAYON
$A6  constant  SSD1306_NORMALDISPLAY   \ default, 1 turns pixel on, 0 turns pixel off
$A7  constant  SSD1306_INVERTDISPLAY   \ inverted logic (active low): 0 turns pixel on, etc.
$81  constant  SSD1306_SETCONTRAST     \ takes 1 parameter (0-255), higher value increases contrast draws more current, $7F = default
$8D  constant  SSD1306_CHARGEPUMP
$20  constant  SSD1306_MEMORYMODE      
  $00  constant  SSD1306_MEMORY_COL+
  $01  constant  SSD1306_MEMORY_ROW+
  $10  constant  SSD1306_MEMORY_PAGE+
$C0  constant  SSD1306_COMSCANINC      \ rowscan direction (commons remap)
$C8  constant  SSD1306_COMSCANDEC      
$A0  constant  SSD1306_SEGSCANINC     \ colscan direction (segments remap)
$A1  constant  SSD1306_SEGSCANDEC
$22  constant  SSD1306_ROWRANGE       \ row/page/common addr
$21  constant  SSD1306_COLRANGE       \ col/segment     addr
$10  constant  SSD1306_SETHICOL       \ page addressing mode
$00  constant  SSD1306_SETLOCOL       \ page addressing mode
$B0  constant  SSD1306_SETROW         \ page start address

\ display on/off
: oled-sleep SSD1306_DISPLAYOFF 1 oled-cmd ;  \ default
: oled-wake  SSD1306_DISPLAYON 1 oled-cmd ;

\ charge pump
: oled-dim    SSD1306_CHARGEPUMP $10 2 oled-cmd ; \ disable (default)
: oled-bright SSD1306_CHARGEPUMP $14 2 oled-cmd ; \ enable

\ contrast control 256-steps ($FF brightest, $0 dimmest), brighter consumes more current
: oled-brightness ( level -- )  SSD1306_SETCONTRAST  swap  2 oled-cmd ;   \ default $8F

\ display mode
: oled-normal  SSD1306_NORMALDISPLAY 1 oled-cmd ;
: oled-invert  SSD1306_INVERTDISPLAY 1 oled-cmd ;

\ memory address modes - first two are text mode, last is graphics mode
: oled-page+  SSD1306_MEMORYMODE  SSD1306_MEMORY_PAGE+ 2 oled-cmd ;  \ auto-advance to right, but loop in same row (page)  default
: oled-row+   SSD1306_MEMORYMODE  SSD1306_MEMORY_ROW+ 2 oled-cmd ;  \ vertical increment - auto-advance down, then right
: oled-col+   SSD1306_MEMORYMODE  SSD1306_MEMORY_COL+ 2 oled-cmd ;  \ horizontal increment - auto-advance to right then down, re-start at top left

\ screen orientation (scanning setup) - allows installing display in any layout.   screen flips on the fly
: oled-pins-top       SSD1306_COMSCANDEC  SSD1306_SEGSCANDEC  2 oled-cmd   oled-col+ ;  
: oled-pins-bottom    SSD1306_COMSCANINC  SSD1306_SEGSCANINC  2 oled-cmd   oled-col+ ; 
: oled-pins-left      SSD1306_COMSCANDEC  SSD1306_SEGSCANINC  2 oled-cmd   oled-row+ ;
: oled-pins-right     SSD1306_COMSCANINC  SSD1306_SEGSCANDEC  2 oled-cmd   oled-row+ ;
\ TODO: in left/right mode, need the pixel expansion, instead of row expansion (only 8 rows)
\ TODO: add dynamic flip of display and correction of image on screen using an orientation sensor

\ window navigation
: oled-window-full ( -- ) \ full screen, init cursor to top left.  restricts pixel changes to window range: row=[0,7], col= [0,$7F]
   SSD1306_ROWRANGE 0 7   SSD1306_COLRANGE 0 $7F  6 oled-cmd ;  
: oled-rowstart    ( -- )   oled-page+  $10 $00 2 oled-cmd oled-col+ ; \ Reset col to 0 at start of current row
: oled-colselect   ( n -- ) oled-page+  SSD1306_SETHICOL /mod  
          SSD1306_SETHICOL or   swap    SSD1306_SETLOCOL or   2 oled-cmd oled-col+ ;  \ selects column
: oled-rowselect   ( n -- p ) 8 /mod    SSD1306_SETROW   or   1 oled-cmd ;  \ calculates which row/page, returns which bit in page

\ pixel graphics - NOTE: clobbers other pixels in same page (row of 8 pixels)
: set-pixel ( p -- ) 8 mod 1 swap lshift 1 oled-data ;  \ [0-7] pth pixel from top of colbyte, does wrap around 
: next-page  oled-row+ $0 1 oled-data  oled-col+ ;  \ In oled-col+ mode, cursor advances to next col, in row+ mode, to the next row
: .cr next-page oled-rowstart ;
: set8   ( p -- ) 8 0 do dup set-pixel loop drop ;
: set64  ( p -- ) 8 0 do dup set8 loop drop ;
: set128 ( p -- ) 2 0 do dup set64 loop drop ;
: diag8  (   -- ) 8 0 do i set-pixel loop ;
: diag64 (   -- ) 8 0 do diag8 next-page loop ;

: set-pos  ( r c -- p ) oled-colselect oled-rowselect ;
: draw-dot ( r c -- )  set-pos set-pixel ;  \ given screen coordinates [0-63]x[0-127]
: clear-row ( r c -- ) dup -rot set-pos drop 128 swap do 0 1 oled-data loop ;

: clear-line ( r c -- ) dup -rot set-pos 128 rot do dup 1 oled-data loop drop ; \ TODO: Can't do this properly without READ


\ TODO: draw Turtle
\ TODO: Pacman
\ TODO: animation, snowflakes falling, drawing fractals, screensavers, or any graphical element...
\ TODO - read colbyte GDDRAM memory so can mask the additional pixel on/off or not (vs. create/save display buffer)

\ NOTE: this command is not super useful.  would be better to lock display with all pixels off
: oled-fill-lock   $A5 1 oled-cmd ;  \ install fill with display locked (GDDRAM can be changed in background) until unlock
: oled-fill-unlock $A4 1 oled-cmd ;

16 constant MAX_I2C_PACKET  \ 32 is sometimes glitchy on Atmega328P  \ TODO: increase speed of I2C xfer to get visual snappyness
: [oled-fillpacket] ( asc -- asc )  MAX_I2C_PACKET 0    do dup loop    MAX_I2C_PACKET oled-data ;   
: [oled-fillrow]    ( asc -- asc )  128 MAX_I2C_PACKET / 0 do [oled-fillpacket] loop ;
\ FEATURE: store pins top/bottom state, and clear screen by flipping memory mode row+/col+ to get a nice smooth scrool
\ may not be as relevant when clearing directly from uC (faster)
: [oled-fill]  ( asc -- )  oled-window-full oled-row+ 8 0 do  [oled-fillrow]  loop  oled-col+ drop ;
: oled-clear-slow  $00 [oled-fill] ;  \ slow-function
: oled-fill-slow   $FF [oled-fill] ;  \ slow-function
: oled-clear  $00 xstore CALL_I2C_OLEDFILL xcall ;   \ fast-function  -- no need for 500ms delay with I2C speed of 10,000
: oled-fill   $FF xstore CALL_I2C_OLEDFILL xcall ;   \ fast-function


: oled-image ( index -- )  xstore CALL_I2C_OLEDIMAGE xcall 500 ms ; \ fast-function, reads from flash memory

: oled-char   ( ascii -- )  xstore CALL_I2C_OLEDCHAR xcall ;
: oled-string ( sptr len -- ) 0 do c@+ oled-char loop drop ;  \  print string  Ex: s" hello" oled-string.  Much faster than ..oled
: .oled  oled-char ;
: ..oled oled-string ;

: oled-init  \ using INIT sequence embedded into target
  i2c-init
  1 i2c-setclock                    \ setting: 1 is snappy, 10 is slow
  OLED_I2C_ADDRESS  oled-setreg
  CALL_I2C_OLEDINIT xcall
  oled-fill
  250 ms
  oled-clear
;

: oled-init-slow  \ sending INIT sequence from host - so can experiment with variations
  i2c-init
  1 i2c-setclock  \ FIX - 1 is snappy, 10 is slow
  OLED_I2C_ADDRESS oled-setreg
  oled-bright
  $8F oled-brightness
  oled-pins-top     
  oled-window-full
  oled-fill
  oled-wake
  250 ms
  oled-clear
;

: oled-photobook  
oled-clear
s" TINY PHOTO!" oled-string .cr
s" by Assad Ebrahim" oled-string .cr
s" www.mathscitech.org" oled-string .cr
s" /articles/TinyPhoto" oled-string .cr
3000 ms
begin 7 2 do oled-clear i oled-image 1500 ms loop key? until 
;


: menu-fast \ works well with yellow-blue OLED screens 128x64
oled-clear
0 10 set-pos s" BREAKFAST" oled-string
16 20 set-pos s" A. Coco Pops" oled-string
24 20 set-pos s" B. Rice Krispies" oled-string
32 20 set-pos s" C. Egg McMuffin" oled-string
40 20 set-pos s" D. Alpha Bites" oled-string
56 10 set-pos s" Select option" oled-string
;



: oled-init1 ( -- )
    $D5 $80  \ osc. frequency controlling frame frequency
    $A8 63   \ set rows (multiplex)
    $DA $12  \ set COM pins configuration
    $D9 $F1  \ set per-charge period
    8 oled-cmd ;
    
: oled-init2 ( -- )
    $D3 $0    \ start row (offset)
    $40 $0 or \ memory map start row
    3 oled-cmd ;

: oled-init6 ( -- )
    $DB $40 $A4 $A6 $2E $AF  6 oled-cmd ;


\ ==== APPLICATIONS ====
: .note   14   .oled ;
: .rest   22   .oled ;
: .space  32   .oled ;
: .minus  45   .oled ; 
: .rarrow 16   .oled ;
: .oarrow 111  .oled ;
: .larrow 17   .oled ;
: .plus   43   .oled ; 
: .cross  88   .oled ;
: .star   42   .oled ;

\ SOFTWARE MENUS: 
\ - < o > +
\ ESC < OK > +
\ BACK < OK > +
\ X < TICK > +
: .menu5  oled-init  .minus .space .larrow .space .oarrow .space .rarrow .space .plus .cr  ;

\ SOFTWARE DICE:
\ need random number roller
\ APPLICATION: D&D / Fighting Fantasy accompaniment...
: .dice6  oled-clear .cr .cr          s" * * *"     ..oled  .cr .cr                  .cr .cr s" * * *" ..oled                      ;
: .dice5  oled-clear .cr .cr          s" *   *"     ..oled  .cr .cr s"   *  " ..oled .cr .cr s" *   *" ..oled ;
: .dice4  oled-clear .cr .cr          s" *   *"     ..oled  .cr .cr                  .cr .cr s" *   *" ..oled                      ;
: .dice3  oled-clear .cr .cr          s" *"         ..oled  .cr .cr s"   *"   ..oled .cr .cr s"     *" ..oled ;
: .dice2  oled-clear .cr .cr                                .cr .cr s" *   *" ..oled .cr .cr ;
: .dice1  oled-clear .cr .cr                                .cr .cr s"   *"   ..oled .cr .cr ;
create .dice ' .dice1 , ' .dice2 , ' .dice3 , ' .dice4 , ' .dice5 , ' .dice6 ,
: dice-show ( n -- ) 1 - cells .dice + @ execute ;

\ ====================================
\ GRAPHICS

\ EXERCISE - Animated square moving across and down the screen
\ TODO: keys input -- controlling the ball
\ TODO: random perturbations --- computer trying too swing the sphere, you need to control it, timer --- mobius the game :)

: top>b8 ( b1 b2 b3 b4 b5 b6 b7 b8 -- vb ) \ produce vertical byte with b1 as lsb
    8 nrev drop >b8 ;

: 4x4sqA ( r c -- ) 
  set-pos drop
  $00 3 0 do dup loop 
  $0F 3 0 do dup loop 
  8 oled-data ;

: 4x4sqB ( r c -- ) 
  set-pos drop
  $0F 3 0 do dup loop 
  $00 3 0 do dup loop 
   8 oled-data ;
  
: 4x4sqC ( r c -- )
  set-pos drop
  $F0 3 0 do dup loop 
  $00 3 0 do dup loop 
   8 oled-data ;

: 4x4sqD ( r c -- )
  set-pos drop
  $00 3 0 do dup loop 
  $F0 3 0 do dup loop 
   8 oled-data ;
   
variable magic#
: magic! ( n -- ) magic# ! ;
: magic@ ( -- n ) magic# @ ;
: magic++ ( -- # ) magic@ 3 + dup magic! ;
: reset-magic ( r c -- r c ) 2dup swap 8 / 128 * + magic# ! ;
: magic-pos ( r c -- r' c' ) 2drop magic++ 128 /mod 8 * 64 mod swap ;

: 4x4sq ( r c -- )  \ animation
  oled-clear
  reset-magic
  begin 
    2dup 4x4sqA 10 ms
    2dup 4x4sqB 10 ms
    2dup 4x4sqC 10 ms
    2dup 4x4sqD 10 ms
    magic-pos     \ slow drift along columns and rows
\    cr .s key drop
  key?
  until
;

\ Graphic - tiles, 6x6 is a grid
: 8x8sq
  $FF 7 0 do dup loop 8 oled-data ;
: 6x6sq
  $0 $7E 5 0 do dup loop $0 8 oled-data ;
: tile oled-clear 128 0 do 6x6sq loop ;

\ Example GForth keyboard control loop with drawing program   
: getkey ( -- k) ekey $FF and ;  \ k in {0,1,2,3}  In GForth, key does not give escaped characters properly, so use ekey
: lt ( 0 ." left"  )  1 -   128 mod         0 ; \ col-
: rt ( 1 ." right" )  1 +   128 mod         0 ; \ col+
: up ( 2 ." up"    ) swap 8 - 64 mod swap   0 ; \ row-
: dn ( 3 ." down"  ) swap 8 + 64 mod swap   0 ; \ row+
: hm ( 4 ." home"  ) 2drop 0 0              0 ; \ go home
: ed ( 5 ." end"   ) 2drop 56 120           0 ; \ go end
: pu ( 6 ) ." pen-up"  0 ; 
: pd ( 7 ) ." pen-down"  0 ; 
: bs ( 8 ) ." backspace"  0 ; 
: de ( 9 ) ." delete"  0 ;     
create response ' lt , ' rt , ' up , ' dn , ' hm , ' ed , ' pu , ' pd , ' bs , ' de ,
: say-esc ." esc" .s cr  oled-clear  ;

: move-ball ( r c -- r' c' )
    2dup set-pos drop 8x8sq 
    getkey dup 0 9 between? 
        if cells response + @ execute 
        else esc-key
            if say-esc 0 else -1 then 
        then 
    >r .s r> cr ; 
: [oled-draw] ( r c -- r' c' )  oled-clear  cr ( ." start" ) .s cr begin move-ball until ;    
: oled-draw ( -- )  0 0 [oled-draw] 2drop ;
\ TODO: Add delete
\ TODO: Add pen-up
\ TODO: Add read GDDRAM and save for re-creation (as an image)
\ TODO: Move to fine brush, fat brush, i.e. 8x8 -> 4x4 -> 2x2 -> 1x1...
\ DONE - Add home and end


\ ====================================

: help-i2c
  cr
  cr ." Help I2C (help-i2c)"
  cr
  cr ." include   3iForth.fs    3iI2C.fs    [3iFont.fs]   3iArt.fs"
  cr
  cr ." 1 open-com   rx"
  cr
  cr ." oled-init"
  cr 
  cr ." Low level:"
  cr ." ---------"
  cr ." oled-sleep  oled-wake"
  cr ." oled-bright  oled-dim"
  cr ." [0-255] oled-brightness"
  cr ." oled-fill  oled-clear" 
  cr ." oled-normal  oled-invert"
  cr ." oled-pins-top  oled-pins-bottom"
  cr
  cr ." High level:"
  cr ." ----------"
  cr ." [ r, c]   set-pos  draw-dot"
  cr ." 11 oled-char"
  cr ." squote Hello!"
  cr ." oled-photobook"
  cr
;

help-i2c


\ need I2C communication primitives (see Wire.h/.cpp)
\ primitive is Arduino's setClock() function which sets clock speed / crystal
\ safe default for I2C is 100 kHz (standard speed) or 400 kHz (fast).  Nano clock speed is 16MHz.
\ CPU frequency is scaled so it cannot generate all possible frequencies

\ : setclk
\    TWBR = ((F_CPU / frequency) - 16) / 2;  \ sets register  .e.g. 16000 / 100 ->72  - note this uses TWI (two-wire interface)

\ ==========================================

( Writing this by translating the driver software into Forth.  )
( There is so much to discard in C! C code feels like obfuscation-ware )

\ Version History:
\ V0 - 2021-06-03---Thu---14:24  - started
\    - 2021-06-08---Tue---13:13 - revisiting following completion of 3iI2C.avr
\ V1 - 2021-06-09---Wed---04:27  - WORKS! with revised 3iForth.fs comms protocol
\ V2 - 2021-07-20---Tue---01:25 - completed set-pos, draw-dot, and clear-row capabilities
\ V3 - 2021-07-21---Wed---06:08 - FIXED snappy display problem -- set I2C_SPEED to 10k
\ V4 - 2021-07-24---Sat---02:42 - ADDED embedded font and fast character display 
\ V5 - 2021-07-25---Sun---01:00 - ADDED moving 4x4 sq animation.  ADDED drawing program with 8x8 squares and GForth key presses...


\ Next Steps:
\   + Embedded turtle logo Forth!
\   + Joystick or key or button control with an OLED I2C display
\   + If use TFT screen can actually do it with a TFT display
\   + positioning cursor arbitrarily > row+ writing rotated/unrotated.  Pixel activation...
\   + lf and cr (newline)
\   + easy display scrolling

\   + Graphics: how to set the ROW?  
\   + How to read the RAM to mask (set pixel) without clobbering other pixels?
\       Maybe it's to set the address to the READ mode, i.e. 0x3D instead of 3C?
\   
\   + Scrolling screen...

\   REFLASH THE FIRMWARE - SIMPLIFIED THE I2C OLED INTERFACE BY REMOVING RESTORECLOCK CMD (forth.avr, i2c.avr, forth.fs, i2c.fs)

\   + Graphics - Piano visual display simulator
\   + Graphics - Pong the game
\   + Addressible LEDs: - https://www.thesmarthomehookup.com/the-complete-guide-to-selecting-individually-addressable-led-strips/


\ include 3iI2C.fs
\ include 3iFont.fs
\ include 3iI2C.fs
\ include 3iArt.fs
\ include 3iArt-butterfly.fs
\ include 3iArt-portraits.fs

