( 3iPianoLights.fs - AE - 2021-05-10---Mon---00:12 )
( This program is for the 32-key LED piano lights board for teaching music theory. )
( The lights board is run from the Arduino Nano running 3iForth.avr code, and run through GForth )
(     include 3iForth.fs, 3iSound-4x4.fs, 3iShift.fs )

\ Hardware: Arduino Nano
\           PianoLights 32 LED board with 19 white LEDs and 13 blue LEDs --- 5VDC, GND
\                       4-pin port on PianoBoard l2r:  DAT(green A2), CS(GND), LAT(yellow A4), CLK(orange A3)
\ Firmware: 3iForth.avr
\ Software & USAGE: Gforth, see help-pianolights below

$1D constant CALL_SHIFTOUT32
: shout32 ( byte-msb byte-3 byte-2 byte-lsb -- )  \ send 4 bytes over serial to target and use embedded shift-out register code
     4 nstore CALL_SHIFTOUT32 xcall ;
: code32  $1E $E $6 $2 ; 
: test32  code32 shout32 ;

\  Design: using 32-bit words in GForth to make this easier
$FF 24 lshift constant maskb4  \ byte MSB
$FF 16 lshift constant maskb3
$FF  8 lshift constant maskb2
$FF           constant maskb1  \ byte LSB
: int2bytes ( int -- byte-msb byte3 byte2 byte-lsb ) \ extract the 4 bytes, lsb top of stack
    >r r@ 24 rshift
    r@ maskb3 and 16 rshift
    r@ maskb2 and 8 rshift
    r> maskb1 and ;

: clr ( -- )  0 int2bytes shout32 ;


\ ------------------------------- LIGHTING ONE LED AT A TIME --- using 8-bit constructs
\   Design: KEY1 = left-most key on board.   KEY32 = right-most key on board
create lightkeymap   \ left to right, white and black keys in order together
    \ 32 bits 
\ keys 1-8  
  $80 c, 0 c, 0 c, 0 c, 
  $40 c, 0 c, 0 c, 0 c, 
  $20 c, 0 c, 0 c, 0 c, 
  $10 c, 0 c, 0 c, 0 c, 
  $08 c, 0 c, 0 c, 0 c, 
  $04 c, 0 c, 0 c, 0 c, 
  $02 c, 0 c, 0 c, 0 c, 
  $01 c, 0 c, 0 c, 0 c, 
\ keys 9-16
  0 c, $80 c, 0 c, 0 c,
  0 c, $40 c, 0 c, 0 c,
  0 c, $20 c, 0 c, 0 c,
  0 c, $10 c, 0 c, 0 c,
  0 c, $08 c, 0 c, 0 c,
  0 c, $04 c, 0 c, 0 c,
  0 c, $02 c, 0 c, 0 c,
  0 c, $01 c, 0 c, 0 c,
\ keys 17-24
  0 c, 0 c, $80 c, 0 c,
  0 c, 0 c, $40 c, 0 c,
  0 c, 0 c, $20 c, 0 c,
  0 c, 0 c, $10 c, 0 c,
  0 c, 0 c, $08 c, 0 c,
  0 c, 0 c, $04 c, 0 c,
  0 c, 0 c, $02 c, 0 c,
  0 c, 0 c, $01 c, 0 c,
\ keys 25-32
  0 c, 0 c, 0 c, $80 c,
  0 c, 0 c, 0 c, $40 c,
  0 c, 0 c, 0 c, $20 c,
  0 c, 0 c, 0 c, $10 c,
  0 c, 0 c, 0 c, $08 c,
  0 c, 0 c, 0 c, $04 c,
  0 c, 0 c, 0 c, $02 c,
  0 c, 0 c, 0 c, $01 c,

: get-key ( index -- <4 bytes> )  \ index: 1-32
    1 - 4 * lightkeymap + 3 0 do c@+ swap loop c@ ;
: light-key ( index -- ) get-key shout32 ;
: kk ( # -- ) light-key ;
: hh disable-output 200 ms enable-output ;  \ flash

\ demo - piano lights board  ---- this forms an EXCELLENT introduction to intermediate forth: memory, data structures, and arrays
\                                 within the powerful context of hardware interfacing/visual programming  (try out with Jai)
\                                 requires Arduino Nano microcontroller with 3iForth.avr, usb cable, and piano lights board 

: kk_  ( # -- ) kk 50 ms ;
: l2r32  33 0 do i kk_ loop ;
: r2l32  31 0 do 31 i - kk_ loop ;
: kit32  begin l2r32 r2l32 key? until 32-clear ;

create blues 2 c, 4 c, 6 c, 9 c, 11 c, 14 c, 16 c, 18 c, 21 c, 23 c, 26 c, 28 c, 30 c, 
create whites 1 c, 3 c, 5 c, 7 c, 8 c, 10 c, 12 c, 13 c, 15 c, 17 c, 19 c, 20 c, 22 c, 24 c, 25 c, 27 c, 29 c, 31 c, 
13 constant num-blues
19 constant num-whites

: color-l2r  ( array narray -- ) swap c@+ rot 0 do kk_ c@+ loop 2drop ;
: color-r2l  ( array narray -- ) byte go-to-end swap c@- rot 0 do kk_ c@- loop 2drop ;
: blues-l2r   blues num-blues color-l2r ; 
: blues-r2l   blues num-blues color-r2l ;
: whites-l2r  whites num-whites color-l2r ; 
: whites-r2l  whites num-whites color-r2l ;

: blues-kit   begin blues-l2r blues-r2l key? until 32-clear ;
: whites-kit  begin whites-l2r whites-r2l key? until 32-clear ;
: rtw-cw      begin blues-l2r whites-r2l key? until 32-clear ;  \ rtw = round the world
: rtw-ccw     begin whites-l2r blues-r2l key? until 32-clear ;

: run-til-key  ( xt -- )  begin dup execute key? until drop key drop ;
: lights-demo  ['] kit32 run-til-key 
               ['] blues-kit run-til-key 
               ['] whites-kit run-til-key 
               ['] rtw-cw run-til-key 
               ['] rtw-ccw run-til-key ;


\ ------------------------------- LIGHTING MULTIPLE LEDS AT A TIME - using 32-bit integers that can be added
create keycode  \ optimized using 32-bit integers on GForth   - integer values (4 bytes) for each key
  1 31 lshift ,  \ ex: int2bytes shout32
  1 30 lshift ,
  1 29 lshift ,
  1 28 lshift ,
  1 27 lshift ,
  1 26 lshift ,
  1 25 lshift ,
  1 24 lshift ,
  1 23 lshift ,
  1 22 lshift ,
  1 21 lshift ,
  1 20 lshift ,
  1 19 lshift ,
  1 18 lshift ,
  1 17 lshift ,
  1 16 lshift ,
  1 15 lshift ,
  1 14 lshift ,
  1 13 lshift ,
  1 12 lshift ,
  1 11 lshift ,
  1 10 lshift ,
  1  9 lshift ,
  1  8 lshift ,
  1  7 lshift ,
  1  6 lshift ,
  1  5 lshift ,
  1  4 lshift ,
  1  3 lshift ,
  1  2 lshift ,
  1  1 lshift ,
  1 ,
\   
: get-code ( index -- ) 1 - cells keycode + @ ;
\ CODE REVIEW: the below does not do well if the codes are duplicates (it still adds them)
: add-codes ( key1..keyn n>1 -- ) >r get-code r> 1 - 0 do swap get-code + loop ;  \ ability to light array of UNIQUE keys (no duplicates) at same time

\ Design:
: lightn ( key1..keyn n>1 -- )  add-codes int2bytes shout32 ;  \ CODE REVIEW: this is called nlight in other modes
: triad  ( a b c -- )     3 lightn ;
: 7th    ( a b c d -- )   4 lightn ;
: 9th    ( a b c d e -- ) 5 lightn ;


\ -------------------------------------------------------



\ Design: 32-key pianolights board HW is an F__ (1) to C'' (32) keyboard, (see 3iSound-4x4.fs, #_  #__ __)
\ Design: For 32-key keyboards F__ to C'' an offset of 8 + is required to get to C (for 32-key pianolights board)
\ Remapping pianolightsboards to notes
\     8 constant cc  \ C is 8th light because lightsboard starts at F__
\    10 constant dd
\    12 constant ee
\    13 constant ff
\    15 constant gg
\    17 constant aa
\    19 constant bb
\    20 constant cc'
\ : %% 1 - ;
\ : ## 1 + ;


  
\ Mapping pianolights board keys Relative to indexed tones in 3iSound-4x4.fs
\ : ., 28 - kk ;   \ #29 is F


\ ------------------------

: help-pianolights ( -- ) 
  cr ." help-pianolights"
  cr ." Help for 32-key Piano Lights Board"
  cr ." ----------------------------------"
  cr
  cr ."    include 3iForth.fs, 3iSound-4x4.fs, 3iShift.fs, 3iPianoLights.fs"
  cr
  cr ."    1 open-com  rx "
  cr ."    init-shiftout    13 start-speaker"
  cr
  cr ."    32-on   32-clear"
  cr ."    lighting numbered LED:              1 light-key   32 light-key  or  8 kk"
  cr ."    lighting pinao notes:               cc dd ee ff gg aa bb cc'   [kk]"
  cr ."    lighting any triple:                8 12 13 triad"
  cr ."    lighting any four-tuple:            8 12 15 19  7th"
  cr ."    lighting any five-tuple:            8 12 15 19 22  9th"
  cr ."    lighting chords based on root:      1 mj    1 mn   1 dm   1 ag    1 sus4   1 sus2  ( chords )"
  cr ."                                        1 dom7  1 mj7  1 mn7  1 hdm7  1 dm7"
  cr ."    transpose chords forward/backward:  chf  chb  ( preserves chord )"
  cr ."    hh  ( flash light )"
  cr 
  cr ."    kit32  blues-kit  whites-kit  rtw-cw  rtw-ccw  (demo)"
  cr ."    lights-demo   ( press key to cycle through )"
;

help-pianolights


\ VERSIONS
\ V1 2021-05-11---Tue---05:05 - ADDED: PianoLights.fs note and chord mappings and major, minor, dim, aug, sus4/sus2 chords
\ V2 2021-05-12---Wed---02:22 - ADDED: 7th and 9th chords, single note names, flash, and full help file
\ V3 2021-05-14---Fri---05:27 - ADDED: lights-demo for the piano lights board
\ V4 2021-05-15---Sat---16:56 - ADDED: Arpeggiator feature, controllable from keys or keyboard, with lights!  Integrated with 3iRecord.fs
\ V5 2021-09-06---Mon---05:31 - FEATURE: Added testchord - for playing and lighting any chord based on sequence
   \ Streamlining for easier usage with Arpeggiator.fs
\  FEATURE: created chordator -- plays chords of specified shape based on pressing one key 
\ (will need good interface for it -- and theory for how to use it)

\ TODO: add scales - all the notes of the scale
\ TODO: play note and light up.   Light up and play note.
